# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from orb import Orb, AsyncOrb
from orb._utils import parse_date, parse_datetime
from tests.utils import assert_matches_type
from orb.pagination import SyncPage, AsyncPage
from orb.types.customers.credits import (
    LedgerListResponse,
    LedgerCreateEntryResponse,
    LedgerListByExternalIDResponse,
    LedgerCreateEntryByExternalIDResponse,
)

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestLedger:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @parametrize
    def test_method_list(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.list(
            customer_id="customer_id",
        )
        assert_matches_type(SyncPage[LedgerListResponse], ledger, path=["response"])

    @parametrize
    def test_method_list_with_all_params(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.list(
            customer_id="customer_id",
            created_at_gt=parse_datetime("2019-12-27T18:11:19.117Z"),
            created_at_gte=parse_datetime("2019-12-27T18:11:19.117Z"),
            created_at_lt=parse_datetime("2019-12-27T18:11:19.117Z"),
            created_at_lte=parse_datetime("2019-12-27T18:11:19.117Z"),
            currency="currency",
            cursor="cursor",
            entry_status="committed",
            entry_type="increment",
            limit=1,
            minimum_amount="minimum_amount",
        )
        assert_matches_type(SyncPage[LedgerListResponse], ledger, path=["response"])

    @parametrize
    def test_raw_response_list(self, client: Orb) -> None:
        response = client.customers.credits.ledger.with_raw_response.list(
            customer_id="customer_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(SyncPage[LedgerListResponse], ledger, path=["response"])

    @parametrize
    def test_streaming_response_list(self, client: Orb) -> None:
        with client.customers.credits.ledger.with_streaming_response.list(
            customer_id="customer_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = response.parse()
            assert_matches_type(SyncPage[LedgerListResponse], ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_list(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            client.customers.credits.ledger.with_raw_response.list(
                customer_id="",
            )

    @parametrize
    def test_method_create_entry_overload_1(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.create_entry(
            customer_id="customer_id",
            amount=0,
            entry_type="increment",
        )
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    def test_method_create_entry_with_all_params_overload_1(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.create_entry(
            customer_id="customer_id",
            amount=0,
            entry_type="increment",
            currency="currency",
            description="description",
            effective_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            expiry_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            invoice_settings={
                "auto_collection": True,
                "net_terms": 0,
                "custom_due_date": parse_date("2019-12-27"),
                "invoice_date": parse_date("2019-12-27"),
                "memo": "memo",
                "require_successful_payment": True,
            },
            metadata={"foo": "string"},
            per_unit_cost_basis="per_unit_cost_basis",
        )
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    def test_raw_response_create_entry_overload_1(self, client: Orb) -> None:
        response = client.customers.credits.ledger.with_raw_response.create_entry(
            customer_id="customer_id",
            amount=0,
            entry_type="increment",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    def test_streaming_response_create_entry_overload_1(self, client: Orb) -> None:
        with client.customers.credits.ledger.with_streaming_response.create_entry(
            customer_id="customer_id",
            amount=0,
            entry_type="increment",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = response.parse()
            assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_create_entry_overload_1(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            client.customers.credits.ledger.with_raw_response.create_entry(
                customer_id="",
                amount=0,
                entry_type="increment",
            )

    @parametrize
    def test_method_create_entry_overload_2(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.create_entry(
            customer_id="customer_id",
            amount=0,
            entry_type="decrement",
        )
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    def test_method_create_entry_with_all_params_overload_2(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.create_entry(
            customer_id="customer_id",
            amount=0,
            entry_type="decrement",
            currency="currency",
            description="description",
            metadata={"foo": "string"},
        )
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    def test_raw_response_create_entry_overload_2(self, client: Orb) -> None:
        response = client.customers.credits.ledger.with_raw_response.create_entry(
            customer_id="customer_id",
            amount=0,
            entry_type="decrement",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    def test_streaming_response_create_entry_overload_2(self, client: Orb) -> None:
        with client.customers.credits.ledger.with_streaming_response.create_entry(
            customer_id="customer_id",
            amount=0,
            entry_type="decrement",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = response.parse()
            assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_create_entry_overload_2(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            client.customers.credits.ledger.with_raw_response.create_entry(
                customer_id="",
                amount=0,
                entry_type="decrement",
            )

    @parametrize
    def test_method_create_entry_overload_3(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.create_entry(
            customer_id="customer_id",
            entry_type="expiration_change",
            target_expiry_date=parse_date("2019-12-27"),
        )
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    def test_method_create_entry_with_all_params_overload_3(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.create_entry(
            customer_id="customer_id",
            entry_type="expiration_change",
            target_expiry_date=parse_date("2019-12-27"),
            amount=0,
            block_id="block_id",
            currency="currency",
            description="description",
            expiry_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            metadata={"foo": "string"},
        )
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    def test_raw_response_create_entry_overload_3(self, client: Orb) -> None:
        response = client.customers.credits.ledger.with_raw_response.create_entry(
            customer_id="customer_id",
            entry_type="expiration_change",
            target_expiry_date=parse_date("2019-12-27"),
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    def test_streaming_response_create_entry_overload_3(self, client: Orb) -> None:
        with client.customers.credits.ledger.with_streaming_response.create_entry(
            customer_id="customer_id",
            entry_type="expiration_change",
            target_expiry_date=parse_date("2019-12-27"),
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = response.parse()
            assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_create_entry_overload_3(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            client.customers.credits.ledger.with_raw_response.create_entry(
                customer_id="",
                entry_type="expiration_change",
                target_expiry_date=parse_date("2019-12-27"),
            )

    @parametrize
    def test_method_create_entry_overload_4(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.create_entry(
            customer_id="customer_id",
            amount=0,
            block_id="block_id",
            entry_type="void",
        )
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    def test_method_create_entry_with_all_params_overload_4(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.create_entry(
            customer_id="customer_id",
            amount=0,
            block_id="block_id",
            entry_type="void",
            currency="currency",
            description="description",
            metadata={"foo": "string"},
            void_reason="refund",
        )
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    def test_raw_response_create_entry_overload_4(self, client: Orb) -> None:
        response = client.customers.credits.ledger.with_raw_response.create_entry(
            customer_id="customer_id",
            amount=0,
            block_id="block_id",
            entry_type="void",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    def test_streaming_response_create_entry_overload_4(self, client: Orb) -> None:
        with client.customers.credits.ledger.with_streaming_response.create_entry(
            customer_id="customer_id",
            amount=0,
            block_id="block_id",
            entry_type="void",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = response.parse()
            assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_create_entry_overload_4(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            client.customers.credits.ledger.with_raw_response.create_entry(
                customer_id="",
                amount=0,
                block_id="block_id",
                entry_type="void",
            )

    @parametrize
    def test_method_create_entry_overload_5(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.create_entry(
            customer_id="customer_id",
            amount=0,
            block_id="block_id",
            entry_type="amendment",
        )
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    def test_method_create_entry_with_all_params_overload_5(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.create_entry(
            customer_id="customer_id",
            amount=0,
            block_id="block_id",
            entry_type="amendment",
            currency="currency",
            description="description",
            metadata={"foo": "string"},
        )
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    def test_raw_response_create_entry_overload_5(self, client: Orb) -> None:
        response = client.customers.credits.ledger.with_raw_response.create_entry(
            customer_id="customer_id",
            amount=0,
            block_id="block_id",
            entry_type="amendment",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    def test_streaming_response_create_entry_overload_5(self, client: Orb) -> None:
        with client.customers.credits.ledger.with_streaming_response.create_entry(
            customer_id="customer_id",
            amount=0,
            block_id="block_id",
            entry_type="amendment",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = response.parse()
            assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_create_entry_overload_5(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            client.customers.credits.ledger.with_raw_response.create_entry(
                customer_id="",
                amount=0,
                block_id="block_id",
                entry_type="amendment",
            )

    @parametrize
    def test_method_create_entry_by_external_id_overload_1(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            entry_type="increment",
        )
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    def test_method_create_entry_by_external_id_with_all_params_overload_1(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            entry_type="increment",
            currency="currency",
            description="description",
            effective_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            expiry_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            invoice_settings={
                "auto_collection": True,
                "net_terms": 0,
                "custom_due_date": parse_date("2019-12-27"),
                "invoice_date": parse_date("2019-12-27"),
                "memo": "memo",
                "require_successful_payment": True,
            },
            metadata={"foo": "string"},
            per_unit_cost_basis="per_unit_cost_basis",
        )
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    def test_raw_response_create_entry_by_external_id_overload_1(self, client: Orb) -> None:
        response = client.customers.credits.ledger.with_raw_response.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            entry_type="increment",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    def test_streaming_response_create_entry_by_external_id_overload_1(self, client: Orb) -> None:
        with client.customers.credits.ledger.with_streaming_response.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            entry_type="increment",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = response.parse()
            assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_create_entry_by_external_id_overload_1(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `external_customer_id` but received ''"):
            client.customers.credits.ledger.with_raw_response.create_entry_by_external_id(
                external_customer_id="",
                amount=0,
                entry_type="increment",
            )

    @parametrize
    def test_method_create_entry_by_external_id_overload_2(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            entry_type="decrement",
        )
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    def test_method_create_entry_by_external_id_with_all_params_overload_2(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            entry_type="decrement",
            currency="currency",
            description="description",
            metadata={"foo": "string"},
        )
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    def test_raw_response_create_entry_by_external_id_overload_2(self, client: Orb) -> None:
        response = client.customers.credits.ledger.with_raw_response.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            entry_type="decrement",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    def test_streaming_response_create_entry_by_external_id_overload_2(self, client: Orb) -> None:
        with client.customers.credits.ledger.with_streaming_response.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            entry_type="decrement",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = response.parse()
            assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_create_entry_by_external_id_overload_2(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `external_customer_id` but received ''"):
            client.customers.credits.ledger.with_raw_response.create_entry_by_external_id(
                external_customer_id="",
                amount=0,
                entry_type="decrement",
            )

    @parametrize
    def test_method_create_entry_by_external_id_overload_3(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            entry_type="expiration_change",
            target_expiry_date=parse_date("2019-12-27"),
        )
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    def test_method_create_entry_by_external_id_with_all_params_overload_3(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            entry_type="expiration_change",
            target_expiry_date=parse_date("2019-12-27"),
            amount=0,
            block_id="block_id",
            currency="currency",
            description="description",
            expiry_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            metadata={"foo": "string"},
        )
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    def test_raw_response_create_entry_by_external_id_overload_3(self, client: Orb) -> None:
        response = client.customers.credits.ledger.with_raw_response.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            entry_type="expiration_change",
            target_expiry_date=parse_date("2019-12-27"),
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    def test_streaming_response_create_entry_by_external_id_overload_3(self, client: Orb) -> None:
        with client.customers.credits.ledger.with_streaming_response.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            entry_type="expiration_change",
            target_expiry_date=parse_date("2019-12-27"),
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = response.parse()
            assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_create_entry_by_external_id_overload_3(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `external_customer_id` but received ''"):
            client.customers.credits.ledger.with_raw_response.create_entry_by_external_id(
                external_customer_id="",
                entry_type="expiration_change",
                target_expiry_date=parse_date("2019-12-27"),
            )

    @parametrize
    def test_method_create_entry_by_external_id_overload_4(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            block_id="block_id",
            entry_type="void",
        )
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    def test_method_create_entry_by_external_id_with_all_params_overload_4(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            block_id="block_id",
            entry_type="void",
            currency="currency",
            description="description",
            metadata={"foo": "string"},
            void_reason="refund",
        )
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    def test_raw_response_create_entry_by_external_id_overload_4(self, client: Orb) -> None:
        response = client.customers.credits.ledger.with_raw_response.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            block_id="block_id",
            entry_type="void",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    def test_streaming_response_create_entry_by_external_id_overload_4(self, client: Orb) -> None:
        with client.customers.credits.ledger.with_streaming_response.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            block_id="block_id",
            entry_type="void",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = response.parse()
            assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_create_entry_by_external_id_overload_4(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `external_customer_id` but received ''"):
            client.customers.credits.ledger.with_raw_response.create_entry_by_external_id(
                external_customer_id="",
                amount=0,
                block_id="block_id",
                entry_type="void",
            )

    @parametrize
    def test_method_create_entry_by_external_id_overload_5(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            block_id="block_id",
            entry_type="amendment",
        )
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    def test_method_create_entry_by_external_id_with_all_params_overload_5(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            block_id="block_id",
            entry_type="amendment",
            currency="currency",
            description="description",
            metadata={"foo": "string"},
        )
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    def test_raw_response_create_entry_by_external_id_overload_5(self, client: Orb) -> None:
        response = client.customers.credits.ledger.with_raw_response.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            block_id="block_id",
            entry_type="amendment",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    def test_streaming_response_create_entry_by_external_id_overload_5(self, client: Orb) -> None:
        with client.customers.credits.ledger.with_streaming_response.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            block_id="block_id",
            entry_type="amendment",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = response.parse()
            assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_create_entry_by_external_id_overload_5(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `external_customer_id` but received ''"):
            client.customers.credits.ledger.with_raw_response.create_entry_by_external_id(
                external_customer_id="",
                amount=0,
                block_id="block_id",
                entry_type="amendment",
            )

    @parametrize
    def test_method_list_by_external_id(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.list_by_external_id(
            external_customer_id="external_customer_id",
        )
        assert_matches_type(SyncPage[LedgerListByExternalIDResponse], ledger, path=["response"])

    @parametrize
    def test_method_list_by_external_id_with_all_params(self, client: Orb) -> None:
        ledger = client.customers.credits.ledger.list_by_external_id(
            external_customer_id="external_customer_id",
            created_at_gt=parse_datetime("2019-12-27T18:11:19.117Z"),
            created_at_gte=parse_datetime("2019-12-27T18:11:19.117Z"),
            created_at_lt=parse_datetime("2019-12-27T18:11:19.117Z"),
            created_at_lte=parse_datetime("2019-12-27T18:11:19.117Z"),
            currency="currency",
            cursor="cursor",
            entry_status="committed",
            entry_type="increment",
            limit=1,
            minimum_amount="minimum_amount",
        )
        assert_matches_type(SyncPage[LedgerListByExternalIDResponse], ledger, path=["response"])

    @parametrize
    def test_raw_response_list_by_external_id(self, client: Orb) -> None:
        response = client.customers.credits.ledger.with_raw_response.list_by_external_id(
            external_customer_id="external_customer_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(SyncPage[LedgerListByExternalIDResponse], ledger, path=["response"])

    @parametrize
    def test_streaming_response_list_by_external_id(self, client: Orb) -> None:
        with client.customers.credits.ledger.with_streaming_response.list_by_external_id(
            external_customer_id="external_customer_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = response.parse()
            assert_matches_type(SyncPage[LedgerListByExternalIDResponse], ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_list_by_external_id(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `external_customer_id` but received ''"):
            client.customers.credits.ledger.with_raw_response.list_by_external_id(
                external_customer_id="",
            )


class TestAsyncLedger:
    parametrize = pytest.mark.parametrize(
        "async_client", [False, True, {"http_client": "aiohttp"}], indirect=True, ids=["loose", "strict", "aiohttp"]
    )

    @parametrize
    async def test_method_list(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.list(
            customer_id="customer_id",
        )
        assert_matches_type(AsyncPage[LedgerListResponse], ledger, path=["response"])

    @parametrize
    async def test_method_list_with_all_params(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.list(
            customer_id="customer_id",
            created_at_gt=parse_datetime("2019-12-27T18:11:19.117Z"),
            created_at_gte=parse_datetime("2019-12-27T18:11:19.117Z"),
            created_at_lt=parse_datetime("2019-12-27T18:11:19.117Z"),
            created_at_lte=parse_datetime("2019-12-27T18:11:19.117Z"),
            currency="currency",
            cursor="cursor",
            entry_status="committed",
            entry_type="increment",
            limit=1,
            minimum_amount="minimum_amount",
        )
        assert_matches_type(AsyncPage[LedgerListResponse], ledger, path=["response"])

    @parametrize
    async def test_raw_response_list(self, async_client: AsyncOrb) -> None:
        response = await async_client.customers.credits.ledger.with_raw_response.list(
            customer_id="customer_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(AsyncPage[LedgerListResponse], ledger, path=["response"])

    @parametrize
    async def test_streaming_response_list(self, async_client: AsyncOrb) -> None:
        async with async_client.customers.credits.ledger.with_streaming_response.list(
            customer_id="customer_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = await response.parse()
            assert_matches_type(AsyncPage[LedgerListResponse], ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_list(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            await async_client.customers.credits.ledger.with_raw_response.list(
                customer_id="",
            )

    @parametrize
    async def test_method_create_entry_overload_1(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.create_entry(
            customer_id="customer_id",
            amount=0,
            entry_type="increment",
        )
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    async def test_method_create_entry_with_all_params_overload_1(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.create_entry(
            customer_id="customer_id",
            amount=0,
            entry_type="increment",
            currency="currency",
            description="description",
            effective_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            expiry_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            invoice_settings={
                "auto_collection": True,
                "net_terms": 0,
                "custom_due_date": parse_date("2019-12-27"),
                "invoice_date": parse_date("2019-12-27"),
                "memo": "memo",
                "require_successful_payment": True,
            },
            metadata={"foo": "string"},
            per_unit_cost_basis="per_unit_cost_basis",
        )
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    async def test_raw_response_create_entry_overload_1(self, async_client: AsyncOrb) -> None:
        response = await async_client.customers.credits.ledger.with_raw_response.create_entry(
            customer_id="customer_id",
            amount=0,
            entry_type="increment",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    async def test_streaming_response_create_entry_overload_1(self, async_client: AsyncOrb) -> None:
        async with async_client.customers.credits.ledger.with_streaming_response.create_entry(
            customer_id="customer_id",
            amount=0,
            entry_type="increment",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = await response.parse()
            assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_create_entry_overload_1(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            await async_client.customers.credits.ledger.with_raw_response.create_entry(
                customer_id="",
                amount=0,
                entry_type="increment",
            )

    @parametrize
    async def test_method_create_entry_overload_2(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.create_entry(
            customer_id="customer_id",
            amount=0,
            entry_type="decrement",
        )
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    async def test_method_create_entry_with_all_params_overload_2(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.create_entry(
            customer_id="customer_id",
            amount=0,
            entry_type="decrement",
            currency="currency",
            description="description",
            metadata={"foo": "string"},
        )
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    async def test_raw_response_create_entry_overload_2(self, async_client: AsyncOrb) -> None:
        response = await async_client.customers.credits.ledger.with_raw_response.create_entry(
            customer_id="customer_id",
            amount=0,
            entry_type="decrement",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    async def test_streaming_response_create_entry_overload_2(self, async_client: AsyncOrb) -> None:
        async with async_client.customers.credits.ledger.with_streaming_response.create_entry(
            customer_id="customer_id",
            amount=0,
            entry_type="decrement",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = await response.parse()
            assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_create_entry_overload_2(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            await async_client.customers.credits.ledger.with_raw_response.create_entry(
                customer_id="",
                amount=0,
                entry_type="decrement",
            )

    @parametrize
    async def test_method_create_entry_overload_3(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.create_entry(
            customer_id="customer_id",
            entry_type="expiration_change",
            target_expiry_date=parse_date("2019-12-27"),
        )
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    async def test_method_create_entry_with_all_params_overload_3(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.create_entry(
            customer_id="customer_id",
            entry_type="expiration_change",
            target_expiry_date=parse_date("2019-12-27"),
            amount=0,
            block_id="block_id",
            currency="currency",
            description="description",
            expiry_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            metadata={"foo": "string"},
        )
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    async def test_raw_response_create_entry_overload_3(self, async_client: AsyncOrb) -> None:
        response = await async_client.customers.credits.ledger.with_raw_response.create_entry(
            customer_id="customer_id",
            entry_type="expiration_change",
            target_expiry_date=parse_date("2019-12-27"),
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    async def test_streaming_response_create_entry_overload_3(self, async_client: AsyncOrb) -> None:
        async with async_client.customers.credits.ledger.with_streaming_response.create_entry(
            customer_id="customer_id",
            entry_type="expiration_change",
            target_expiry_date=parse_date("2019-12-27"),
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = await response.parse()
            assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_create_entry_overload_3(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            await async_client.customers.credits.ledger.with_raw_response.create_entry(
                customer_id="",
                entry_type="expiration_change",
                target_expiry_date=parse_date("2019-12-27"),
            )

    @parametrize
    async def test_method_create_entry_overload_4(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.create_entry(
            customer_id="customer_id",
            amount=0,
            block_id="block_id",
            entry_type="void",
        )
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    async def test_method_create_entry_with_all_params_overload_4(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.create_entry(
            customer_id="customer_id",
            amount=0,
            block_id="block_id",
            entry_type="void",
            currency="currency",
            description="description",
            metadata={"foo": "string"},
            void_reason="refund",
        )
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    async def test_raw_response_create_entry_overload_4(self, async_client: AsyncOrb) -> None:
        response = await async_client.customers.credits.ledger.with_raw_response.create_entry(
            customer_id="customer_id",
            amount=0,
            block_id="block_id",
            entry_type="void",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    async def test_streaming_response_create_entry_overload_4(self, async_client: AsyncOrb) -> None:
        async with async_client.customers.credits.ledger.with_streaming_response.create_entry(
            customer_id="customer_id",
            amount=0,
            block_id="block_id",
            entry_type="void",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = await response.parse()
            assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_create_entry_overload_4(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            await async_client.customers.credits.ledger.with_raw_response.create_entry(
                customer_id="",
                amount=0,
                block_id="block_id",
                entry_type="void",
            )

    @parametrize
    async def test_method_create_entry_overload_5(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.create_entry(
            customer_id="customer_id",
            amount=0,
            block_id="block_id",
            entry_type="amendment",
        )
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    async def test_method_create_entry_with_all_params_overload_5(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.create_entry(
            customer_id="customer_id",
            amount=0,
            block_id="block_id",
            entry_type="amendment",
            currency="currency",
            description="description",
            metadata={"foo": "string"},
        )
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    async def test_raw_response_create_entry_overload_5(self, async_client: AsyncOrb) -> None:
        response = await async_client.customers.credits.ledger.with_raw_response.create_entry(
            customer_id="customer_id",
            amount=0,
            block_id="block_id",
            entry_type="amendment",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

    @parametrize
    async def test_streaming_response_create_entry_overload_5(self, async_client: AsyncOrb) -> None:
        async with async_client.customers.credits.ledger.with_streaming_response.create_entry(
            customer_id="customer_id",
            amount=0,
            block_id="block_id",
            entry_type="amendment",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = await response.parse()
            assert_matches_type(LedgerCreateEntryResponse, ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_create_entry_overload_5(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            await async_client.customers.credits.ledger.with_raw_response.create_entry(
                customer_id="",
                amount=0,
                block_id="block_id",
                entry_type="amendment",
            )

    @parametrize
    async def test_method_create_entry_by_external_id_overload_1(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            entry_type="increment",
        )
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    async def test_method_create_entry_by_external_id_with_all_params_overload_1(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            entry_type="increment",
            currency="currency",
            description="description",
            effective_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            expiry_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            invoice_settings={
                "auto_collection": True,
                "net_terms": 0,
                "custom_due_date": parse_date("2019-12-27"),
                "invoice_date": parse_date("2019-12-27"),
                "memo": "memo",
                "require_successful_payment": True,
            },
            metadata={"foo": "string"},
            per_unit_cost_basis="per_unit_cost_basis",
        )
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    async def test_raw_response_create_entry_by_external_id_overload_1(self, async_client: AsyncOrb) -> None:
        response = await async_client.customers.credits.ledger.with_raw_response.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            entry_type="increment",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    async def test_streaming_response_create_entry_by_external_id_overload_1(self, async_client: AsyncOrb) -> None:
        async with async_client.customers.credits.ledger.with_streaming_response.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            entry_type="increment",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = await response.parse()
            assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_create_entry_by_external_id_overload_1(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `external_customer_id` but received ''"):
            await async_client.customers.credits.ledger.with_raw_response.create_entry_by_external_id(
                external_customer_id="",
                amount=0,
                entry_type="increment",
            )

    @parametrize
    async def test_method_create_entry_by_external_id_overload_2(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            entry_type="decrement",
        )
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    async def test_method_create_entry_by_external_id_with_all_params_overload_2(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            entry_type="decrement",
            currency="currency",
            description="description",
            metadata={"foo": "string"},
        )
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    async def test_raw_response_create_entry_by_external_id_overload_2(self, async_client: AsyncOrb) -> None:
        response = await async_client.customers.credits.ledger.with_raw_response.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            entry_type="decrement",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    async def test_streaming_response_create_entry_by_external_id_overload_2(self, async_client: AsyncOrb) -> None:
        async with async_client.customers.credits.ledger.with_streaming_response.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            entry_type="decrement",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = await response.parse()
            assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_create_entry_by_external_id_overload_2(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `external_customer_id` but received ''"):
            await async_client.customers.credits.ledger.with_raw_response.create_entry_by_external_id(
                external_customer_id="",
                amount=0,
                entry_type="decrement",
            )

    @parametrize
    async def test_method_create_entry_by_external_id_overload_3(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            entry_type="expiration_change",
            target_expiry_date=parse_date("2019-12-27"),
        )
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    async def test_method_create_entry_by_external_id_with_all_params_overload_3(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            entry_type="expiration_change",
            target_expiry_date=parse_date("2019-12-27"),
            amount=0,
            block_id="block_id",
            currency="currency",
            description="description",
            expiry_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            metadata={"foo": "string"},
        )
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    async def test_raw_response_create_entry_by_external_id_overload_3(self, async_client: AsyncOrb) -> None:
        response = await async_client.customers.credits.ledger.with_raw_response.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            entry_type="expiration_change",
            target_expiry_date=parse_date("2019-12-27"),
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    async def test_streaming_response_create_entry_by_external_id_overload_3(self, async_client: AsyncOrb) -> None:
        async with async_client.customers.credits.ledger.with_streaming_response.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            entry_type="expiration_change",
            target_expiry_date=parse_date("2019-12-27"),
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = await response.parse()
            assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_create_entry_by_external_id_overload_3(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `external_customer_id` but received ''"):
            await async_client.customers.credits.ledger.with_raw_response.create_entry_by_external_id(
                external_customer_id="",
                entry_type="expiration_change",
                target_expiry_date=parse_date("2019-12-27"),
            )

    @parametrize
    async def test_method_create_entry_by_external_id_overload_4(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            block_id="block_id",
            entry_type="void",
        )
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    async def test_method_create_entry_by_external_id_with_all_params_overload_4(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            block_id="block_id",
            entry_type="void",
            currency="currency",
            description="description",
            metadata={"foo": "string"},
            void_reason="refund",
        )
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    async def test_raw_response_create_entry_by_external_id_overload_4(self, async_client: AsyncOrb) -> None:
        response = await async_client.customers.credits.ledger.with_raw_response.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            block_id="block_id",
            entry_type="void",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    async def test_streaming_response_create_entry_by_external_id_overload_4(self, async_client: AsyncOrb) -> None:
        async with async_client.customers.credits.ledger.with_streaming_response.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            block_id="block_id",
            entry_type="void",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = await response.parse()
            assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_create_entry_by_external_id_overload_4(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `external_customer_id` but received ''"):
            await async_client.customers.credits.ledger.with_raw_response.create_entry_by_external_id(
                external_customer_id="",
                amount=0,
                block_id="block_id",
                entry_type="void",
            )

    @parametrize
    async def test_method_create_entry_by_external_id_overload_5(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            block_id="block_id",
            entry_type="amendment",
        )
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    async def test_method_create_entry_by_external_id_with_all_params_overload_5(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            block_id="block_id",
            entry_type="amendment",
            currency="currency",
            description="description",
            metadata={"foo": "string"},
        )
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    async def test_raw_response_create_entry_by_external_id_overload_5(self, async_client: AsyncOrb) -> None:
        response = await async_client.customers.credits.ledger.with_raw_response.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            block_id="block_id",
            entry_type="amendment",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

    @parametrize
    async def test_streaming_response_create_entry_by_external_id_overload_5(self, async_client: AsyncOrb) -> None:
        async with async_client.customers.credits.ledger.with_streaming_response.create_entry_by_external_id(
            external_customer_id="external_customer_id",
            amount=0,
            block_id="block_id",
            entry_type="amendment",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = await response.parse()
            assert_matches_type(LedgerCreateEntryByExternalIDResponse, ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_create_entry_by_external_id_overload_5(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `external_customer_id` but received ''"):
            await async_client.customers.credits.ledger.with_raw_response.create_entry_by_external_id(
                external_customer_id="",
                amount=0,
                block_id="block_id",
                entry_type="amendment",
            )

    @parametrize
    async def test_method_list_by_external_id(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.list_by_external_id(
            external_customer_id="external_customer_id",
        )
        assert_matches_type(AsyncPage[LedgerListByExternalIDResponse], ledger, path=["response"])

    @parametrize
    async def test_method_list_by_external_id_with_all_params(self, async_client: AsyncOrb) -> None:
        ledger = await async_client.customers.credits.ledger.list_by_external_id(
            external_customer_id="external_customer_id",
            created_at_gt=parse_datetime("2019-12-27T18:11:19.117Z"),
            created_at_gte=parse_datetime("2019-12-27T18:11:19.117Z"),
            created_at_lt=parse_datetime("2019-12-27T18:11:19.117Z"),
            created_at_lte=parse_datetime("2019-12-27T18:11:19.117Z"),
            currency="currency",
            cursor="cursor",
            entry_status="committed",
            entry_type="increment",
            limit=1,
            minimum_amount="minimum_amount",
        )
        assert_matches_type(AsyncPage[LedgerListByExternalIDResponse], ledger, path=["response"])

    @parametrize
    async def test_raw_response_list_by_external_id(self, async_client: AsyncOrb) -> None:
        response = await async_client.customers.credits.ledger.with_raw_response.list_by_external_id(
            external_customer_id="external_customer_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        ledger = response.parse()
        assert_matches_type(AsyncPage[LedgerListByExternalIDResponse], ledger, path=["response"])

    @parametrize
    async def test_streaming_response_list_by_external_id(self, async_client: AsyncOrb) -> None:
        async with async_client.customers.credits.ledger.with_streaming_response.list_by_external_id(
            external_customer_id="external_customer_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            ledger = await response.parse()
            assert_matches_type(AsyncPage[LedgerListByExternalIDResponse], ledger, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_list_by_external_id(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `external_customer_id` but received ''"):
            await async_client.customers.credits.ledger.with_raw_response.list_by_external_id(
                external_customer_id="",
            )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from orb import Orb, AsyncOrb
from tests.utils import assert_matches_type
from orb.types.shared import Price

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestExternalPriceID:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @parametrize
    def test_method_update(self, client: Orb) -> None:
        external_price_id = client.prices.external_price_id.update(
            external_price_id="external_price_id",
        )
        assert_matches_type(Price, external_price_id, path=["response"])

    @parametrize
    def test_method_update_with_all_params(self, client: Orb) -> None:
        external_price_id = client.prices.external_price_id.update(
            external_price_id="external_price_id",
            metadata={"foo": "string"},
        )
        assert_matches_type(Price, external_price_id, path=["response"])

    @parametrize
    def test_raw_response_update(self, client: Orb) -> None:
        response = client.prices.external_price_id.with_raw_response.update(
            external_price_id="external_price_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        external_price_id = response.parse()
        assert_matches_type(Price, external_price_id, path=["response"])

    @parametrize
    def test_streaming_response_update(self, client: Orb) -> None:
        with client.prices.external_price_id.with_streaming_response.update(
            external_price_id="external_price_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            external_price_id = response.parse()
            assert_matches_type(Price, external_price_id, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_update(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `external_price_id` but received ''"):
            client.prices.external_price_id.with_raw_response.update(
                external_price_id="",
            )

    @parametrize
    def test_method_fetch(self, client: Orb) -> None:
        external_price_id = client.prices.external_price_id.fetch(
            "external_price_id",
        )
        assert_matches_type(Price, external_price_id, path=["response"])

    @parametrize
    def test_raw_response_fetch(self, client: Orb) -> None:
        response = client.prices.external_price_id.with_raw_response.fetch(
            "external_price_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        external_price_id = response.parse()
        assert_matches_type(Price, external_price_id, path=["response"])

    @parametrize
    def test_streaming_response_fetch(self, client: Orb) -> None:
        with client.prices.external_price_id.with_streaming_response.fetch(
            "external_price_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            external_price_id = response.parse()
            assert_matches_type(Price, external_price_id, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_fetch(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `external_price_id` but received ''"):
            client.prices.external_price_id.with_raw_response.fetch(
                "",
            )


class TestAsyncExternalPriceID:
    parametrize = pytest.mark.parametrize(
        "async_client", [False, True, {"http_client": "aiohttp"}], indirect=True, ids=["loose", "strict", "aiohttp"]
    )

    @parametrize
    async def test_method_update(self, async_client: AsyncOrb) -> None:
        external_price_id = await async_client.prices.external_price_id.update(
            external_price_id="external_price_id",
        )
        assert_matches_type(Price, external_price_id, path=["response"])

    @parametrize
    async def test_method_update_with_all_params(self, async_client: AsyncOrb) -> None:
        external_price_id = await async_client.prices.external_price_id.update(
            external_price_id="external_price_id",
            metadata={"foo": "string"},
        )
        assert_matches_type(Price, external_price_id, path=["response"])

    @parametrize
    async def test_raw_response_update(self, async_client: AsyncOrb) -> None:
        response = await async_client.prices.external_price_id.with_raw_response.update(
            external_price_id="external_price_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        external_price_id = response.parse()
        assert_matches_type(Price, external_price_id, path=["response"])

    @parametrize
    async def test_streaming_response_update(self, async_client: AsyncOrb) -> None:
        async with async_client.prices.external_price_id.with_streaming_response.update(
            external_price_id="external_price_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            external_price_id = await response.parse()
            assert_matches_type(Price, external_price_id, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_update(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `external_price_id` but received ''"):
            await async_client.prices.external_price_id.with_raw_response.update(
                external_price_id="",
            )

    @parametrize
    async def test_method_fetch(self, async_client: AsyncOrb) -> None:
        external_price_id = await async_client.prices.external_price_id.fetch(
            "external_price_id",
        )
        assert_matches_type(Price, external_price_id, path=["response"])

    @parametrize
    async def test_raw_response_fetch(self, async_client: AsyncOrb) -> None:
        response = await async_client.prices.external_price_id.with_raw_response.fetch(
            "external_price_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        external_price_id = response.parse()
        assert_matches_type(Price, external_price_id, path=["response"])

    @parametrize
    async def test_streaming_response_fetch(self, async_client: AsyncOrb) -> None:
        async with async_client.prices.external_price_id.with_streaming_response.fetch(
            "external_price_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            external_price_id = await response.parse()
            assert_matches_type(Price, external_price_id, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_fetch(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `external_price_id` but received ''"):
            await async_client.prices.external_price_id.with_raw_response.fetch(
                "",
            )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional

from .price import Price
from ..._models import BaseModel

__all__ = ["PerPriceCost"]


class PerPriceCost(BaseModel):
    price: Price
    """The price object"""

    price_id: str
    """The price the cost is associated with"""

    subtotal: str
    """Price's contributions for the timeframe, excluding any minimums and discounts."""

    total: str
    """Price's contributions for the timeframe, including minimums and discounts."""

    quantity: Optional[float] = None
    """The price's quantity for the timeframe"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from .dimensional_price_groups import DimensionalPriceGroups as DimensionalPriceGroups
from .external_dimensional_price_group_id_update_params import (
    ExternalDimensionalPriceGroupIDUpdateParams as ExternalDimensionalPriceGroupIDUpdateParams,
)



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Optional

import httpx

from ... import _legacy_response
from ..._types import Body, Omit, Query, Headers, NotGiven, omit, not_given
from ..._utils import maybe_transform, async_maybe_transform
from ..._compat import cached_property
from ..._resource import SyncAPIResource, AsyncAPIResource
from ..._response import to_streamed_response_wrapper, async_to_streamed_response_wrapper
from ...types.plan import Plan
from ...types.plans import external_plan_id_update_params
from ..._base_client import make_request_options

__all__ = ["ExternalPlanID", "AsyncExternalPlanID"]


class ExternalPlanID(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> ExternalPlanIDWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/orbcorp/orb-python#accessing-raw-response-data-eg-headers
        """
        return ExternalPlanIDWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> ExternalPlanIDWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/orbcorp/orb-python#with_streaming_response
        """
        return ExternalPlanIDWithStreamingResponse(self)

    def update(
        self,
        other_external_plan_id: str,
        *,
        external_plan_id: Optional[str] | Omit = omit,
        metadata: Optional[Dict[str, Optional[str]]] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> Plan:
        """
        This endpoint can be used to update the `external_plan_id`, and `metadata` of an
        existing plan.

        Other fields on a plan are currently immutable.

        Args:
          external_plan_id: An optional user-defined ID for this plan resource, used throughout the system
              as an alias for this Plan. Use this field to identify a plan by an existing
              identifier in your system.

          metadata: User-specified key/value pairs for the resource. Individual keys can be removed
              by setting the value to `null`, and the entire metadata mapping can be cleared
              by setting `metadata` to `null`.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not other_external_plan_id:
            raise ValueError(
                f"Expected a non-empty value for `other_external_plan_id` but received {other_external_plan_id!r}"
            )
        return self._put(
            f"/plans/external_plan_id/{other_external_plan_id}",
            body=maybe_transform(
                {
                    "external_plan_id": external_plan_id,
                    "metadata": metadata,
                },
                external_plan_id_update_params.ExternalPlanIDUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=Plan,
        )

    def fetch(
        self,
        external_plan_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> Plan:
        """
        This endpoint is used to fetch [plan](/core-concepts##plan-and-price) details
        given an external_plan_id identifier. It returns information about the prices
        included in the plan and their configuration, as well as the product that the
        plan is attached to.

        If multiple plans are found to contain the specified external_plan_id, the
        active plans will take priority over archived ones, and among those, the
        endpoint will return the most recently created plan.

        ## Serialized prices

        Orb supports a few different pricing models out of the box. Each of these models
        is serialized differently in a given [Price](/core-concepts#plan-and-price)
        object. The `model_type` field determines the key for the configuration object
        that is present. A detailed explanation of price types can be found in the
        [Price schema](/core-concepts#plan-and-price). "

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not external_plan_id:
            raise ValueError(f"Expected a non-empty value for `external_plan_id` but received {external_plan_id!r}")
        return self._get(
            f"/plans/external_plan_id/{external_plan_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=Plan,
        )


class AsyncExternalPlanID(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncExternalPlanIDWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/orbcorp/orb-python#accessing-raw-response-data-eg-headers
        """
        return AsyncExternalPlanIDWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncExternalPlanIDWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/orbcorp/orb-python#with_streaming_response
        """
        return AsyncExternalPlanIDWithStreamingResponse(self)

    async def update(
        self,
        other_external_plan_id: str,
        *,
        external_plan_id: Optional[str] | Omit = omit,
        metadata: Optional[Dict[str, Optional[str]]] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> Plan:
        """
        This endpoint can be used to update the `external_plan_id`, and `metadata` of an
        existing plan.

        Other fields on a plan are currently immutable.

        Args:
          external_plan_id: An optional user-defined ID for this plan resource, used throughout the system
              as an alias for this Plan. Use this field to identify a plan by an existing
              identifier in your system.

          metadata: User-specified key/value pairs for the resource. Individual keys can be removed
              by setting the value to `null`, and the entire metadata mapping can be cleared
              by setting `metadata` to `null`.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not other_external_plan_id:
            raise ValueError(
                f"Expected a non-empty value for `other_external_plan_id` but received {other_external_plan_id!r}"
            )
        return await self._put(
            f"/plans/external_plan_id/{other_external_plan_id}",
            body=await async_maybe_transform(
                {
                    "external_plan_id": external_plan_id,
                    "metadata": metadata,
                },
                external_plan_id_update_params.ExternalPlanIDUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=Plan,
        )

    async def fetch(
        self,
        external_plan_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> Plan:
        """
        This endpoint is used to fetch [plan](/core-concepts##plan-and-price) details
        given an external_plan_id identifier. It returns information about the prices
        included in the plan and their configuration, as well as the product that the
        plan is attached to.

        If multiple plans are found to contain the specified external_plan_id, the
        active plans will take priority over archived ones, and among those, the
        endpoint will return the most recently created plan.

        ## Serialized prices

        Orb supports a few different pricing models out of the box. Each of these models
        is serialized differently in a given [Price](/core-concepts#plan-and-price)
        object. The `model_type` field determines the key for the configuration object
        that is present. A detailed explanation of price types can be found in the
        [Price schema](/core-concepts#plan-and-price). "

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not external_plan_id:
            raise ValueError(f"Expected a non-empty value for `external_plan_id` but received {external_plan_id!r}")
        return await self._get(
            f"/plans/external_plan_id/{external_plan_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=Plan,
        )


class ExternalPlanIDWithRawResponse:
    def __init__(self, external_plan_id: ExternalPlanID) -> None:
        self._external_plan_id = external_plan_id

        self.update = _legacy_response.to_raw_response_wrapper(
            external_plan_id.update,
        )
        self.fetch = _legacy_response.to_raw_response_wrapper(
            external_plan_id.fetch,
        )


class AsyncExternalPlanIDWithRawResponse:
    def __init__(self, external_plan_id: AsyncExternalPlanID) -> None:
        self._external_plan_id = external_plan_id

        self.update = _legacy_response.async_to_raw_response_wrapper(
            external_plan_id.update,
        )
        self.fetch = _legacy_response.async_to_raw_response_wrapper(
            external_plan_id.fetch,
        )


class ExternalPlanIDWithStreamingResponse:
    def __init__(self, external_plan_id: ExternalPlanID) -> None:
        self._external_plan_id = external_plan_id

        self.update = to_streamed_response_wrapper(
            external_plan_id.update,
        )
        self.fetch = to_streamed_response_wrapper(
            external_plan_id.fetch,
        )


class AsyncExternalPlanIDWithStreamingResponse:
    def __init__(self, external_plan_id: AsyncExternalPlanID) -> None:
        self._external_plan_id = external_plan_id

        self.update = async_to_streamed_response_wrapper(
            external_plan_id.update,
        )
        self.fetch = async_to_streamed_response_wrapper(
            external_plan_id.fetch,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union, Optional
from datetime import datetime
from typing_extensions import Literal

import httpx

from ... import _legacy_response
from ..._types import Body, Omit, Query, Headers, NotGiven, omit, not_given
from ..._utils import maybe_transform, async_maybe_transform
from ..._compat import cached_property
from ..._resource import SyncAPIResource, AsyncAPIResource
from ..._response import to_streamed_response_wrapper, async_to_streamed_response_wrapper
from ..._base_client import make_request_options
from ...types.customers import cost_list_params, cost_list_by_external_id_params
from ...types.customers.cost_list_response import CostListResponse
from ...types.customers.cost_list_by_external_id_response import CostListByExternalIDResponse

__all__ = ["Costs", "AsyncCosts"]


class Costs(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> CostsWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/orbcorp/orb-python#accessing-raw-response-data-eg-headers
        """
        return CostsWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> CostsWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/orbcorp/orb-python#with_streaming_response
        """
        return CostsWithStreamingResponse(self)

    def list(
        self,
        customer_id: str,
        *,
        currency: Optional[str] | Omit = omit,
        timeframe_end: Union[str, datetime, None] | Omit = omit,
        timeframe_start: Union[str, datetime, None] | Omit = omit,
        view_mode: Optional[Literal["periodic", "cumulative"]] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> CostListResponse:
        """
        This endpoint is used to fetch a day-by-day snapshot of a customer's costs in
        Orb, calculated by applying pricing information to the underlying usage (see the
        [subscription usage endpoint](/api-reference/subscription/fetch-subscription-usage)
        to fetch usage per metric, in usage units rather than a currency).

        This endpoint can be leveraged for internal tooling and to provide a more
        transparent billing experience for your end users:

        1. Understand the cost breakdown per line item historically and in real-time for
           the current billing period.
        2. Provide customer visibility into how different services are contributing to
           the overall invoice with a per-day timeseries (as compared to the
           [upcoming invoice](/api-reference/invoice/fetch-upcoming-invoice) resource,
           which represents a snapshot for the current period).
        3. Assess how minimums and discounts affect your customers by teasing apart
           costs directly as a result of usage, as opposed to minimums and discounts at
           the plan and price level.
        4. Gain insight into key customer health metrics, such as the percent
           utilization of the minimum committed spend.

        ## Fetching subscriptions

        By default, this endpoint fetches the currently active subscription for the
        customer, and returns cost information for the subscription's current billing
        period, broken down by each participating price. If there are no currently
        active subscriptions, this will instead default to the most recently active
        subscription or return an empty series if none are found. For example, if your
        plan charges for compute hours, job runs, and data syncs, then this endpoint
        would provide a daily breakdown of your customer's cost for each of those axes.

        If timeframe bounds are specified, Orb fetches all subscriptions that were
        active in that timeframe. If two subscriptions overlap on a single day, costs
        from each price will be summed, and prices for both subscriptions will be
        included in the breakdown.

        ## Prepaid plans

        For plans that include prices which deduct credits rather than accrue in-arrears
        charges in a billable currency, this endpoint will return the total deduction
        amount, in credits, for the specified timeframe.

        ## Cumulative subtotals and totals

        Since the subtotal and total must factor in any billing-period level discounts
        and minimums, it's most meaningful to consider costs relative to the start of
        the subscription's billing period. As a result, by default this endpoint returns
        cumulative totals since the beginning of the billing period. In particular, the
        `timeframe_start` of a returned timeframe window is _always_ the beginning of
        the billing period and `timeframe_end` is incremented one day at a time to build
        the result.

        A customer that uses a few API calls a day but has a minimum commitment might
        exhibit the following pattern for their subtotal and total in the first few days
        of the month. Here, we assume that each API call is \\$$2.50, the customer's plan
        has a monthly minimum of \\$$50 for this price, and that the subscription's
        billing period bounds are aligned to the first of the month:

        | timeframe_start | timeframe_end | Cumulative usage | Subtotal | Total (incl. commitment) |
        | --------------- | ------------- | ---------------- | -------- | ------------------------ |
        | 2023-02-01      | 2023-02-02    | 9                | \\$$22.50  | \\$$50.00                  |
        | 2023-02-01      | 2023-02-03    | 19               | \\$$47.50  | \\$$50.00                  |
        | 2023-02-01      | 2023-02-04    | 20               | \\$$50.00  | \\$$50.00                  |
        | 2023-02-01      | 2023-02-05    | 28               | \\$$70.00  | \\$$70.00                  |
        | 2023-02-01      | 2023-02-06    | 36               | \\$$90.00  | \\$$90.00                  |

        ### Periodic values

        When the query parameter `view_mode=periodic` is specified, Orb will return an
        incremental day-by-day view of costs. In this case, there will always be a
        one-day difference between `timeframe_start` and `timeframe_end` for the
        timeframes returned. This is a transform on top of the cumulative costs,
        calculated by taking the difference of each timeframe with the last. Note that
        in the above example, the `Total` value would be 0 for the second two data
        points, since the minimum commitment has not yet been hit and each day is not
        contributing anything to the total cost.

        ## Timeframe bounds

        For an active subscription, both timeframes should be specified in the request.
        If a subscription starts or ends within the timeframe, the response will only
        include windows where the subscription is active. If a subscription has ended,
        no timeframe bounds need to be specified and the response will default to the
        billing period when the subscription was last active.

        As noted above, `timeframe_start` for a given cumulative datapoint is always the
        beginning of the billing period, and `timeframe_end` is incremented one day at a
        time to construct the response. When a timeframe is passed in that is not
        aligned to the current subscription's billing period, the response will contain
        cumulative totals from multiple billing periods.

        Suppose the queried customer has a subscription aligned to the 15th of every
        month. If this endpoint is queried with the date range `2023-06-01` -
        `2023-07-01`, the first data point will represent about half a billing period's
        worth of costs, accounting for accruals from the start of the billing period and
        inclusive of the first day of the timeframe
        (`timeframe_start = 2023-05-15 00:00:00`, `timeframe_end = 2023-06-02 00:00:00`)

        | datapoint index | timeframe_start | timeframe_end |
        | --------------- | --------------- | ------------- |
        | 0               | 2023-05-15      | 2023-06-02    |
        | 1               | 2023-05-15      | 2023-06-03    |
        | 2               | ...             | ...           |
        | 3               | 2023-05-15      | 2023-06-14    |
        | 4               | 2023-06-15      | 2023-06-16    |
        | 5               | 2023-06-15      | 2023-06-17    |
        | 6               | ...             | ...           |
        | 7               | 2023-06-15      | 2023-07-01    |

        You can see this sliced timeframe visualized
        [here](https://i.imgur.com/TXhYgme.png).

        ### Matrix prices

        When a price uses matrix pricing, it's important to view costs grouped by those
        matrix dimensions. Orb will return `price_groups` with the `grouping_key` and
        `secondary_grouping_key` based on the matrix price definition, for each
        `grouping_value` and `secondary_grouping_value` available.

        Args:
          currency: The currency or custom pricing unit to use.

          timeframe_end: Costs returned are exclusive of `timeframe_end`.

          timeframe_start: Costs returned are inclusive of `timeframe_start`.

          view_mode: Controls whether Orb returns cumulative costs since the start of the billing
              period, or incremental day-by-day costs. If your customer has minimums or
              discounts, it's strongly recommended that you use the default cumulative
              behavior.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return self._get(
            f"/customers/{customer_id}/costs",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "currency": currency,
                        "timeframe_end": timeframe_end,
                        "timeframe_start": timeframe_start,
                        "view_mode": view_mode,
                    },
                    cost_list_params.CostListParams,
                ),
            ),
            cast_to=CostListResponse,
        )

    def list_by_external_id(
        self,
        external_customer_id: str,
        *,
        currency: Optional[str] | Omit = omit,
        timeframe_end: Union[str, datetime, None] | Omit = omit,
        timeframe_start: Union[str, datetime, None] | Omit = omit,
        view_mode: Optional[Literal["periodic", "cumulative"]] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> CostListByExternalIDResponse:
        """
        This endpoint is used to fetch a day-by-day snapshot of a customer's costs in
        Orb, calculated by applying pricing information to the underlying usage (see the
        [subscription usage endpoint](/api-reference/subscription/fetch-subscription-usage)
        to fetch usage per metric, in usage units rather than a currency).

        This endpoint can be leveraged for internal tooling and to provide a more
        transparent billing experience for your end users:

        1. Understand the cost breakdown per line item historically and in real-time for
           the current billing period.
        2. Provide customer visibility into how different services are contributing to
           the overall invoice with a per-day timeseries (as compared to the
           [upcoming invoice](/api-reference/invoice/fetch-upcoming-invoice) resource,
           which represents a snapshot for the current period).
        3. Assess how minimums and discounts affect your customers by teasing apart
           costs directly as a result of usage, as opposed to minimums and discounts at
           the plan and price level.
        4. Gain insight into key customer health metrics, such as the percent
           utilization of the minimum committed spend.

        ## Fetching subscriptions

        By default, this endpoint fetches the currently active subscription for the
        customer, and returns cost information for the subscription's current billing
        period, broken down by each participating price. If there are no currently
        active subscriptions, this will instead default to the most recently active
        subscription or return an empty series if none are found. For example, if your
        plan charges for compute hours, job runs, and data syncs, then this endpoint
        would provide a daily breakdown of your customer's cost for each of those axes.

        If timeframe bounds are specified, Orb fetches all subscriptions that were
        active in that timeframe. If two subscriptions overlap on a single day, costs
        from each price will be summed, and prices for both subscriptions will be
        included in the breakdown.

        ## Prepaid plans

        For plans that include prices which deduct credits rather than accrue in-arrears
        charges in a billable currency, this endpoint will return the total deduction
        amount, in credits, for the specified timeframe.

        ## Cumulative subtotals and totals

        Since the subtotal and total must factor in any billing-period level discounts
        and minimums, it's most meaningful to consider costs relative to the start of
        the subscription's billing period. As a result, by default this endpoint returns
        cumulative totals since the beginning of the billing period. In particular, the
        `timeframe_start` of a returned timeframe window is _always_ the beginning of
        the billing period and `timeframe_end` is incremented one day at a time to build
        the result.

        A customer that uses a few API calls a day but has a minimum commitment might
        exhibit the following pattern for their subtotal and total in the first few days
        of the month. Here, we assume that each API call is \\$$2.50, the customer's plan
        has a monthly minimum of \\$$50 for this price, and that the subscription's
        billing period bounds are aligned to the first of the month:

        | timeframe_start | timeframe_end | Cumulative usage | Subtotal | Total (incl. commitment) |
        | --------------- | ------------- | ---------------- | -------- | ------------------------ |
        | 2023-02-01      | 2023-02-02    | 9                | \\$$22.50  | \\$$50.00                  |
        | 2023-02-01      | 2023-02-03    | 19               | \\$$47.50  | \\$$50.00                  |
        | 2023-02-01      | 2023-02-04    | 20               | \\$$50.00  | \\$$50.00                  |
        | 2023-02-01      | 2023-02-05    | 28               | \\$$70.00  | \\$$70.00                  |
        | 2023-02-01      | 2023-02-06    | 36               | \\$$90.00  | \\$$90.00                  |

        ### Periodic values

        When the query parameter `view_mode=periodic` is specified, Orb will return an
        incremental day-by-day view of costs. In this case, there will always be a
        one-day difference between `timeframe_start` and `timeframe_end` for the
        timeframes returned. This is a transform on top of the cumulative costs,
        calculated by taking the difference of each timeframe with the last. Note that
        in the above example, the `Total` value would be 0 for the second two data
        points, since the minimum commitment has not yet been hit and each day is not
        contributing anything to the total cost.

        ## Timeframe bounds

        For an active subscription, both timeframes should be specified in the request.
        If a subscription starts or ends within the timeframe, the response will only
        include windows where the subscription is active. If a subscription has ended,
        no timeframe bounds need to be specified and the response will default to the
        billing period when the subscription was last active.

        As noted above, `timeframe_start` for a given cumulative datapoint is always the
        beginning of the billing period, and `timeframe_end` is incremented one day at a
        time to construct the response. When a timeframe is passed in that is not
        aligned to the current subscription's billing period, the response will contain
        cumulative totals from multiple billing periods.

        Suppose the queried customer has a subscription aligned to the 15th of every
        month. If this endpoint is queried with the date range `2023-06-01` -
        `2023-07-01`, the first data point will represent about half a billing period's
        worth of costs, accounting for accruals from the start of the billing period and
        inclusive of the first day of the timeframe
        (`timeframe_start = 2023-05-15 00:00:00`, `timeframe_end = 2023-06-02 00:00:00`)

        | datapoint index | timeframe_start | timeframe_end |
        | --------------- | --------------- | ------------- |
        | 0               | 2023-05-15      | 2023-06-02    |
        | 1               | 2023-05-15      | 2023-06-03    |
        | 2               | ...             | ...           |
        | 3               | 2023-05-15      | 2023-06-14    |
        | 4               | 2023-06-15      | 2023-06-16    |
        | 5               | 2023-06-15      | 2023-06-17    |
        | 6               | ...             | ...           |
        | 7               | 2023-06-15      | 2023-07-01    |

        You can see this sliced timeframe visualized
        [here](https://i.imgur.com/TXhYgme.png).

        ### Matrix prices

        When a price uses matrix pricing, it's important to view costs grouped by those
        matrix dimensions. Orb will return `price_groups` with the `grouping_key` and
        `secondary_grouping_key` based on the matrix price definition, for each
        `grouping_value` and `secondary_grouping_value` available.

        Args:
          currency: The currency or custom pricing unit to use.

          timeframe_end: Costs returned are exclusive of `timeframe_end`.

          timeframe_start: Costs returned are inclusive of `timeframe_start`.

          view_mode: Controls whether Orb returns cumulative costs since the start of the billing
              period, or incremental day-by-day costs. If your customer has minimums or
              discounts, it's strongly recommended that you use the default cumulative
              behavior.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not external_customer_id:
            raise ValueError(
                f"Expected a non-empty value for `external_customer_id` but received {external_customer_id!r}"
            )
        return self._get(
            f"/customers/external_customer_id/{external_customer_id}/costs",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "currency": currency,
                        "timeframe_end": timeframe_end,
                        "timeframe_start": timeframe_start,
                        "view_mode": view_mode,
                    },
                    cost_list_by_external_id_params.CostListByExternalIDParams,
                ),
            ),
            cast_to=CostListByExternalIDResponse,
        )


class AsyncCosts(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncCostsWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/orbcorp/orb-python#accessing-raw-response-data-eg-headers
        """
        return AsyncCostsWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncCostsWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/orbcorp/orb-python#with_streaming_response
        """
        return AsyncCostsWithStreamingResponse(self)

    async def list(
        self,
        customer_id: str,
        *,
        currency: Optional[str] | Omit = omit,
        timeframe_end: Union[str, datetime, None] | Omit = omit,
        timeframe_start: Union[str, datetime, None] | Omit = omit,
        view_mode: Optional[Literal["periodic", "cumulative"]] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> CostListResponse:
        """
        This endpoint is used to fetch a day-by-day snapshot of a customer's costs in
        Orb, calculated by applying pricing information to the underlying usage (see the
        [subscription usage endpoint](/api-reference/subscription/fetch-subscription-usage)
        to fetch usage per metric, in usage units rather than a currency).

        This endpoint can be leveraged for internal tooling and to provide a more
        transparent billing experience for your end users:

        1. Understand the cost breakdown per line item historically and in real-time for
           the current billing period.
        2. Provide customer visibility into how different services are contributing to
           the overall invoice with a per-day timeseries (as compared to the
           [upcoming invoice](/api-reference/invoice/fetch-upcoming-invoice) resource,
           which represents a snapshot for the current period).
        3. Assess how minimums and discounts affect your customers by teasing apart
           costs directly as a result of usage, as opposed to minimums and discounts at
           the plan and price level.
        4. Gain insight into key customer health metrics, such as the percent
           utilization of the minimum committed spend.

        ## Fetching subscriptions

        By default, this endpoint fetches the currently active subscription for the
        customer, and returns cost information for the subscription's current billing
        period, broken down by each participating price. If there are no currently
        active subscriptions, this will instead default to the most recently active
        subscription or return an empty series if none are found. For example, if your
        plan charges for compute hours, job runs, and data syncs, then this endpoint
        would provide a daily breakdown of your customer's cost for each of those axes.

        If timeframe bounds are specified, Orb fetches all subscriptions that were
        active in that timeframe. If two subscriptions overlap on a single day, costs
        from each price will be summed, and prices for both subscriptions will be
        included in the breakdown.

        ## Prepaid plans

        For plans that include prices which deduct credits rather than accrue in-arrears
        charges in a billable currency, this endpoint will return the total deduction
        amount, in credits, for the specified timeframe.

        ## Cumulative subtotals and totals

        Since the subtotal and total must factor in any billing-period level discounts
        and minimums, it's most meaningful to consider costs relative to the start of
        the subscription's billing period. As a result, by default this endpoint returns
        cumulative totals since the beginning of the billing period. In particular, the
        `timeframe_start` of a returned timeframe window is _always_ the beginning of
        the billing period and `timeframe_end` is incremented one day at a time to build
        the result.

        A customer that uses a few API calls a day but has a minimum commitment might
        exhibit the following pattern for their subtotal and total in the first few days
        of the month. Here, we assume that each API call is \\$$2.50, the customer's plan
        has a monthly minimum of \\$$50 for this price, and that the subscription's
        billing period bounds are aligned to the first of the month:

        | timeframe_start | timeframe_end | Cumulative usage | Subtotal | Total (incl. commitment) |
        | --------------- | ------------- | ---------------- | -------- | ------------------------ |
        | 2023-02-01      | 2023-02-02    | 9                | \\$$22.50  | \\$$50.00                  |
        | 2023-02-01      | 2023-02-03    | 19               | \\$$47.50  | \\$$50.00                  |
        | 2023-02-01      | 2023-02-04    | 20               | \\$$50.00  | \\$$50.00                  |
        | 2023-02-01      | 2023-02-05    | 28               | \\$$70.00  | \\$$70.00                  |
        | 2023-02-01      | 2023-02-06    | 36               | \\$$90.00  | \\$$90.00                  |

        ### Periodic values

        When the query parameter `view_mode=periodic` is specified, Orb will return an
        incremental day-by-day view of costs. In this case, there will always be a
        one-day difference between `timeframe_start` and `timeframe_end` for the
        timeframes returned. This is a transform on top of the cumulative costs,
        calculated by taking the difference of each timeframe with the last. Note that
        in the above example, the `Total` value would be 0 for the second two data
        points, since the minimum commitment has not yet been hit and each day is not
        contributing anything to the total cost.

        ## Timeframe bounds

        For an active subscription, both timeframes should be specified in the request.
        If a subscription starts or ends within the timeframe, the response will only
        include windows where the subscription is active. If a subscription has ended,
        no timeframe bounds need to be specified and the response will default to the
        billing period when the subscription was last active.

        As noted above, `timeframe_start` for a given cumulative datapoint is always the
        beginning of the billing period, and `timeframe_end` is incremented one day at a
        time to construct the response. When a timeframe is passed in that is not
        aligned to the current subscription's billing period, the response will contain
        cumulative totals from multiple billing periods.

        Suppose the queried customer has a subscription aligned to the 15th of every
        month. If this endpoint is queried with the date range `2023-06-01` -
        `2023-07-01`, the first data point will represent about half a billing period's
        worth of costs, accounting for accruals from the start of the billing period and
        inclusive of the first day of the timeframe
        (`timeframe_start = 2023-05-15 00:00:00`, `timeframe_end = 2023-06-02 00:00:00`)

        | datapoint index | timeframe_start | timeframe_end |
        | --------------- | --------------- | ------------- |
        | 0               | 2023-05-15      | 2023-06-02    |
        | 1               | 2023-05-15      | 2023-06-03    |
        | 2               | ...             | ...           |
        | 3               | 2023-05-15      | 2023-06-14    |
        | 4               | 2023-06-15      | 2023-06-16    |
        | 5               | 2023-06-15      | 2023-06-17    |
        | 6               | ...             | ...           |
        | 7               | 2023-06-15      | 2023-07-01    |

        You can see this sliced timeframe visualized
        [here](https://i.imgur.com/TXhYgme.png).

        ### Matrix prices

        When a price uses matrix pricing, it's important to view costs grouped by those
        matrix dimensions. Orb will return `price_groups` with the `grouping_key` and
        `secondary_grouping_key` based on the matrix price definition, for each
        `grouping_value` and `secondary_grouping_value` available.

        Args:
          currency: The currency or custom pricing unit to use.

          timeframe_end: Costs returned are exclusive of `timeframe_end`.

          timeframe_start: Costs returned are inclusive of `timeframe_start`.

          view_mode: Controls whether Orb returns cumulative costs since the start of the billing
              period, or incremental day-by-day costs. If your customer has minimums or
              discounts, it's strongly recommended that you use the default cumulative
              behavior.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return await self._get(
            f"/customers/{customer_id}/costs",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=await async_maybe_transform(
                    {
                        "currency": currency,
                        "timeframe_end": timeframe_end,
                        "timeframe_start": timeframe_start,
                        "view_mode": view_mode,
                    },
                    cost_list_params.CostListParams,
                ),
            ),
            cast_to=CostListResponse,
        )

    async def list_by_external_id(
        self,
        external_customer_id: str,
        *,
        currency: Optional[str] | Omit = omit,
        timeframe_end: Union[str, datetime, None] | Omit = omit,
        timeframe_start: Union[str, datetime, None] | Omit = omit,
        view_mode: Optional[Literal["periodic", "cumulative"]] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> CostListByExternalIDResponse:
        """
        This endpoint is used to fetch a day-by-day snapshot of a customer's costs in
        Orb, calculated by applying pricing information to the underlying usage (see the
        [subscription usage endpoint](/api-reference/subscription/fetch-subscription-usage)
        to fetch usage per metric, in usage units rather than a currency).

        This endpoint can be leveraged for internal tooling and to provide a more
        transparent billing experience for your end users:

        1. Understand the cost breakdown per line item historically and in real-time for
           the current billing period.
        2. Provide customer visibility into how different services are contributing to
           the overall invoice with a per-day timeseries (as compared to the
           [upcoming invoice](/api-reference/invoice/fetch-upcoming-invoice) resource,
           which represents a snapshot for the current period).
        3. Assess how minimums and discounts affect your customers by teasing apart
           costs directly as a result of usage, as opposed to minimums and discounts at
           the plan and price level.
        4. Gain insight into key customer health metrics, such as the percent
           utilization of the minimum committed spend.

        ## Fetching subscriptions

        By default, this endpoint fetches the currently active subscription for the
        customer, and returns cost information for the subscription's current billing
        period, broken down by each participating price. If there are no currently
        active subscriptions, this will instead default to the most recently active
        subscription or return an empty series if none are found. For example, if your
        plan charges for compute hours, job runs, and data syncs, then this endpoint
        would provide a daily breakdown of your customer's cost for each of those axes.

        If timeframe bounds are specified, Orb fetches all subscriptions that were
        active in that timeframe. If two subscriptions overlap on a single day, costs
        from each price will be summed, and prices for both subscriptions will be
        included in the breakdown.

        ## Prepaid plans

        For plans that include prices which deduct credits rather than accrue in-arrears
        charges in a billable currency, this endpoint will return the total deduction
        amount, in credits, for the specified timeframe.

        ## Cumulative subtotals and totals

        Since the subtotal and total must factor in any billing-period level discounts
        and minimums, it's most meaningful to consider costs relative to the start of
        the subscription's billing period. As a result, by default this endpoint returns
        cumulative totals since the beginning of the billing period. In particular, the
        `timeframe_start` of a returned timeframe window is _always_ the beginning of
        the billing period and `timeframe_end` is incremented one day at a time to build
        the result.

        A customer that uses a few API calls a day but has a minimum commitment might
        exhibit the following pattern for their subtotal and total in the first few days
        of the month. Here, we assume that each API call is \\$$2.50, the customer's plan
        has a monthly minimum of \\$$50 for this price, and that the subscription's
        billing period bounds are aligned to the first of the month:

        | timeframe_start | timeframe_end | Cumulative usage | Subtotal | Total (incl. commitment) |
        | --------------- | ------------- | ---------------- | -------- | ------------------------ |
        | 2023-02-01      | 2023-02-02    | 9                | \\$$22.50  | \\$$50.00                  |
        | 2023-02-01      | 2023-02-03    | 19               | \\$$47.50  | \\$$50.00                  |
        | 2023-02-01      | 2023-02-04    | 20               | \\$$50.00  | \\$$50.00                  |
        | 2023-02-01      | 2023-02-05    | 28               | \\$$70.00  | \\$$70.00                  |
        | 2023-02-01      | 2023-02-06    | 36               | \\$$90.00  | \\$$90.00                  |

        ### Periodic values

        When the query parameter `view_mode=periodic` is specified, Orb will return an
        incremental day-by-day view of costs. In this case, there will always be a
        one-day difference between `timeframe_start` and `timeframe_end` for the
        timeframes returned. This is a transform on top of the cumulative costs,
        calculated by taking the difference of each timeframe with the last. Note that
        in the above example, the `Total` value would be 0 for the second two data
        points, since the minimum commitment has not yet been hit and each day is not
        contributing anything to the total cost.

        ## Timeframe bounds

        For an active subscription, both timeframes should be specified in the request.
        If a subscription starts or ends within the timeframe, the response will only
        include windows where the subscription is active. If a subscription has ended,
        no timeframe bounds need to be specified and the response will default to the
        billing period when the subscription was last active.

        As noted above, `timeframe_start` for a given cumulative datapoint is always the
        beginning of the billing period, and `timeframe_end` is incremented one day at a
        time to construct the response. When a timeframe is passed in that is not
        aligned to the current subscription's billing period, the response will contain
        cumulative totals from multiple billing periods.

        Suppose the queried customer has a subscription aligned to the 15th of every
        month. If this endpoint is queried with the date range `2023-06-01` -
        `2023-07-01`, the first data point will represent about half a billing period's
        worth of costs, accounting for accruals from the start of the billing period and
        inclusive of the first day of the timeframe
        (`timeframe_start = 2023-05-15 00:00:00`, `timeframe_end = 2023-06-02 00:00:00`)

        | datapoint index | timeframe_start | timeframe_end |
        | --------------- | --------------- | ------------- |
        | 0               | 2023-05-15      | 2023-06-02    |
        | 1               | 2023-05-15      | 2023-06-03    |
        | 2               | ...             | ...           |
        | 3               | 2023-05-15      | 2023-06-14    |
        | 4               | 2023-06-15      | 2023-06-16    |
        | 5               | 2023-06-15      | 2023-06-17    |
        | 6               | ...             | ...           |
        | 7               | 2023-06-15      | 2023-07-01    |

        You can see this sliced timeframe visualized
        [here](https://i.imgur.com/TXhYgme.png).

        ### Matrix prices

        When a price uses matrix pricing, it's important to view costs grouped by those
        matrix dimensions. Orb will return `price_groups` with the `grouping_key` and
        `secondary_grouping_key` based on the matrix price definition, for each
        `grouping_value` and `secondary_grouping_value` available.

        Args:
          currency: The currency or custom pricing unit to use.

          timeframe_end: Costs returned are exclusive of `timeframe_end`.

          timeframe_start: Costs returned are inclusive of `timeframe_start`.

          view_mode: Controls whether Orb returns cumulative costs since the start of the billing
              period, or incremental day-by-day costs. If your customer has minimums or
              discounts, it's strongly recommended that you use the default cumulative
              behavior.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not external_customer_id:
            raise ValueError(
                f"Expected a non-empty value for `external_customer_id` but received {external_customer_id!r}"
            )
        return await self._get(
            f"/customers/external_customer_id/{external_customer_id}/costs",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=await async_maybe_transform(
                    {
                        "currency": currency,
                        "timeframe_end": timeframe_end,
                        "timeframe_start": timeframe_start,
                        "view_mode": view_mode,
                    },
                    cost_list_by_external_id_params.CostListByExternalIDParams,
                ),
            ),
            cast_to=CostListByExternalIDResponse,
        )


class CostsWithRawResponse:
    def __init__(self, costs: Costs) -> None:
        self._costs = costs

        self.list = _legacy_response.to_raw_response_wrapper(
            costs.list,
        )
        self.list_by_external_id = _legacy_response.to_raw_response_wrapper(
            costs.list_by_external_id,
        )


class AsyncCostsWithRawResponse:
    def __init__(self, costs: AsyncCosts) -> None:
        self._costs = costs

        self.list = _legacy_response.async_to_raw_response_wrapper(
            costs.list,
        )
        self.list_by_external_id = _legacy_response.async_to_raw_response_wrapper(
            costs.list_by_external_id,
        )


class CostsWithStreamingResponse:
    def __init__(self, costs: Costs) -> None:
        self._costs = costs

        self.list = to_streamed_response_wrapper(
            costs.list,
        )
        self.list_by_external_id = to_streamed_response_wrapper(
            costs.list_by_external_id,
        )


class AsyncCostsWithStreamingResponse:
    def __init__(self, costs: AsyncCosts) -> None:
        self._costs = costs

        self.list = async_to_streamed_response_wrapper(
            costs.list,
        )
        self.list_by_external_id = async_to_streamed_response_wrapper(
            costs.list_by_external_id,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .tier import Tier as Tier
from .discount import Discount as Discount
from .bulk_tier import BulkTier as BulkTier
from .bulk_config import BulkConfig as BulkConfig
from .new_maximum import NewMaximum as NewMaximum
from .new_minimum import NewMinimum as NewMinimum
from .unit_config import UnitConfig as UnitConfig
from .matrix_value import MatrixValue as MatrixValue
from .matrix_config import MatrixConfig as MatrixConfig
from .tiered_config import TieredConfig as TieredConfig
from .package_config import PackageConfig as PackageConfig
from .trial_discount import TrialDiscount as TrialDiscount
from .usage_discount import UsageDiscount as UsageDiscount
from .amount_discount import AmountDiscount as AmountDiscount
from .customer_tax_id import CustomerTaxID as CustomerTaxID
from .custom_expiration import CustomExpiration as CustomExpiration
from .new_usage_discount import NewUsageDiscount as NewUsageDiscount
from .new_amount_discount import NewAmountDiscount as NewAmountDiscount
from .new_plan_bulk_price import NewPlanBulkPrice as NewPlanBulkPrice
from .new_plan_unit_price import NewPlanUnitPrice as NewPlanUnitPrice
from .percentage_discount import PercentageDiscount as PercentageDiscount
from .conversion_rate_tier import ConversionRateTier as ConversionRateTier
from .new_allocation_price import NewAllocationPrice as NewAllocationPrice
from .new_plan_matrix_price import NewPlanMatrixPrice as NewPlanMatrixPrice
from .new_plan_tiered_price import NewPlanTieredPrice as NewPlanTieredPrice
from .new_plan_package_price import NewPlanPackagePrice as NewPlanPackagePrice
from .transform_price_filter import TransformPriceFilter as TransformPriceFilter
from .new_floating_bulk_price import NewFloatingBulkPrice as NewFloatingBulkPrice
from .new_floating_unit_price import NewFloatingUnitPrice as NewFloatingUnitPrice
from .new_percentage_discount import NewPercentageDiscount as NewPercentageDiscount
from .new_floating_matrix_price import NewFloatingMatrixPrice as NewFloatingMatrixPrice
from .new_floating_tiered_price import NewFloatingTieredPrice as NewFloatingTieredPrice
from .new_floating_package_price import NewFloatingPackagePrice as NewFloatingPackagePrice
from .billing_cycle_relative_date import BillingCycleRelativeDate as BillingCycleRelativeDate
from .conversion_rate_unit_config import ConversionRateUnitConfig as ConversionRateUnitConfig
from .unit_conversion_rate_config import UnitConversionRateConfig as UnitConversionRateConfig
from .conversion_rate_tiered_config import ConversionRateTieredConfig as ConversionRateTieredConfig
from .matrix_with_allocation_config import MatrixWithAllocationConfig as MatrixWithAllocationConfig
from .new_plan_grouped_tiered_price import NewPlanGroupedTieredPrice as NewPlanGroupedTieredPrice
from .new_plan_tiered_package_price import NewPlanTieredPackagePrice as NewPlanTieredPackagePrice
from .tiered_conversion_rate_config import TieredConversionRateConfig as TieredConversionRateConfig
from .new_billing_cycle_configuration import NewBillingCycleConfiguration as NewBillingCycleConfiguration
from .new_plan_minimum_composite_price import NewPlanMinimumCompositePrice as NewPlanMinimumCompositePrice
from .new_plan_unit_with_percent_price import NewPlanUnitWithPercentPrice as NewPlanUnitWithPercentPrice
from .new_floating_grouped_tiered_price import NewFloatingGroupedTieredPrice as NewFloatingGroupedTieredPrice
from .new_floating_tiered_package_price import NewFloatingTieredPackagePrice as NewFloatingTieredPackagePrice
from .new_plan_grouped_allocation_price import NewPlanGroupedAllocationPrice as NewPlanGroupedAllocationPrice
from .billing_cycle_anchor_configuration import BillingCycleAnchorConfiguration as BillingCycleAnchorConfiguration
from .new_plan_bulk_with_proration_price import NewPlanBulkWithProrationPrice as NewPlanBulkWithProrationPrice
from .new_plan_tiered_with_minimum_price import NewPlanTieredWithMinimumPrice as NewPlanTieredWithMinimumPrice
from .new_plan_unit_with_proration_price import NewPlanUnitWithProrationPrice as NewPlanUnitWithProrationPrice
from .new_dimensional_price_configuration import NewDimensionalPriceConfiguration as NewDimensionalPriceConfiguration
from .new_floating_minimum_composite_price import NewFloatingMinimumCompositePrice as NewFloatingMinimumCompositePrice
from .new_floating_unit_with_percent_price import NewFloatingUnitWithPercentPrice as NewFloatingUnitWithPercentPrice
from .new_floating_grouped_allocation_price import (
    NewFloatingGroupedAllocationPrice as NewFloatingGroupedAllocationPrice,
)
from .new_plan_grouped_tiered_package_price import NewPlanGroupedTieredPackagePrice as NewPlanGroupedTieredPackagePrice
from .new_plan_matrix_with_allocation_price import NewPlanMatrixWithAllocationPrice as NewPlanMatrixWithAllocationPrice
from .new_plan_threshold_total_amount_price import NewPlanThresholdTotalAmountPrice as NewPlanThresholdTotalAmountPrice
from .new_floating_bulk_with_proration_price import (
    NewFloatingBulkWithProrationPrice as NewFloatingBulkWithProrationPrice,
)
from .new_floating_tiered_with_minimum_price import (
    NewFloatingTieredWithMinimumPrice as NewFloatingTieredWithMinimumPrice,
)
from .new_floating_unit_with_proration_price import (
    NewFloatingUnitWithProrationPrice as NewFloatingUnitWithProrationPrice,
)
from .new_plan_cumulative_grouped_bulk_price import (
    NewPlanCumulativeGroupedBulkPrice as NewPlanCumulativeGroupedBulkPrice,
)
from .new_plan_package_with_allocation_price import (
    NewPlanPackageWithAllocationPrice as NewPlanPackageWithAllocationPrice,
)
from .new_plan_matrix_with_display_name_price import (
    NewPlanMatrixWithDisplayNamePrice as NewPlanMatrixWithDisplayNamePrice,
)
from .new_plan_max_group_tiered_package_price import (
    NewPlanMaxGroupTieredPackagePrice as NewPlanMaxGroupTieredPackagePrice,
)
from .new_floating_tiered_with_proration_price import (
    NewFloatingTieredWithProrationPrice as NewFloatingTieredWithProrationPrice,
)
from .new_floating_grouped_tiered_package_price import (
    NewFloatingGroupedTieredPackagePrice as NewFloatingGroupedTieredPackagePrice,
)
from .new_floating_matrix_with_allocation_price import (
    NewFloatingMatrixWithAllocationPrice as NewFloatingMatrixWithAllocationPrice,
)
from .new_floating_threshold_total_amount_price import (
    NewFloatingThresholdTotalAmountPrice as NewFloatingThresholdTotalAmountPrice,
)
from .new_floating_cumulative_grouped_bulk_price import (
    NewFloatingCumulativeGroupedBulkPrice as NewFloatingCumulativeGroupedBulkPrice,
)
from .new_floating_package_with_allocation_price import (
    NewFloatingPackageWithAllocationPrice as NewFloatingPackageWithAllocationPrice,
)
from .new_plan_tiered_package_with_minimum_price import (
    NewPlanTieredPackageWithMinimumPrice as NewPlanTieredPackageWithMinimumPrice,
)
from .new_floating_matrix_with_display_name_price import (
    NewFloatingMatrixWithDisplayNamePrice as NewFloatingMatrixWithDisplayNamePrice,
)
from .new_floating_max_group_tiered_package_price import (
    NewFloatingMaxGroupTieredPackagePrice as NewFloatingMaxGroupTieredPackagePrice,
)
from .new_plan_grouped_with_metered_minimum_price import (
    NewPlanGroupedWithMeteredMinimumPrice as NewPlanGroupedWithMeteredMinimumPrice,
)
from .new_plan_grouped_with_prorated_minimum_price import (
    NewPlanGroupedWithProratedMinimumPrice as NewPlanGroupedWithProratedMinimumPrice,
)
from .new_floating_tiered_package_with_minimum_price import (
    NewFloatingTieredPackageWithMinimumPrice as NewFloatingTieredPackageWithMinimumPrice,
)
from .new_floating_grouped_with_metered_minimum_price import (
    NewFloatingGroupedWithMeteredMinimumPrice as NewFloatingGroupedWithMeteredMinimumPrice,
)
from .new_floating_grouped_with_prorated_minimum_price import (
    NewFloatingGroupedWithProratedMinimumPrice as NewFloatingGroupedWithProratedMinimumPrice,
)
from .new_plan_scalable_matrix_with_unit_pricing_price import (
    NewPlanScalableMatrixWithUnitPricingPrice as NewPlanScalableMatrixWithUnitPricingPrice,
)
from .new_plan_scalable_matrix_with_tiered_pricing_price import (
    NewPlanScalableMatrixWithTieredPricingPrice as NewPlanScalableMatrixWithTieredPricingPrice,
)
from .new_floating_scalable_matrix_with_unit_pricing_price import (
    NewFloatingScalableMatrixWithUnitPricingPrice as NewFloatingScalableMatrixWithUnitPricingPrice,
)
from .new_floating_scalable_matrix_with_tiered_pricing_price import (
    NewFloatingScalableMatrixWithTieredPricingPrice as NewFloatingScalableMatrixWithTieredPricingPrice,
)



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Optional
from typing_extensions import Required, TypedDict

from ..._types import SequenceNotStr

__all__ = ["MatrixValue"]


class MatrixValue(TypedDict, total=False):
    dimension_values: Required[SequenceNotStr[Optional[str]]]
    """One or two matrix keys to filter usage to this Matrix value by"""

    unit_amount: Required[str]
    """Unit price for the specified dimension_values"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union, Optional
from datetime import datetime

import httpx

from ... import _legacy_response
from ..._types import Body, Omit, Query, Headers, NotGiven, omit, not_given
from ..._utils import maybe_transform, async_maybe_transform
from ..._compat import cached_property
from ..._resource import SyncAPIResource, AsyncAPIResource
from ..._response import to_streamed_response_wrapper, async_to_streamed_response_wrapper
from ..._base_client import make_request_options
from ...types.events import volume_list_params
from ...types.events.event_volumes import EventVolumes

__all__ = ["Volume", "AsyncVolume"]


class Volume(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> VolumeWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/orbcorp/orb-python#accessing-raw-response-data-eg-headers
        """
        return VolumeWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> VolumeWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/orbcorp/orb-python#with_streaming_response
        """
        return VolumeWithStreamingResponse(self)

    def list(
        self,
        *,
        timeframe_start: Union[str, datetime],
        cursor: Optional[str] | Omit = omit,
        limit: int | Omit = omit,
        timeframe_end: Union[str, datetime] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> EventVolumes:
        """
        This endpoint returns the event volume for an account in a
        [paginated list format](/api-reference/pagination).

        The event volume is aggregated by the hour and the
        [timestamp](/api-reference/event/ingest-events) field is used to determine which
        hour an event is associated with. Note, this means that late-arriving events
        increment the volume count for the hour window the timestamp is in, not the
        latest hour window.

        Each item in the response contains the count of events aggregated by the hour
        where the start and end time are hour-aligned and in UTC. When a specific
        timestamp is passed in for either start or end time, the response includes the
        hours the timestamp falls in.

        Args:
          timeframe_start: The start of the timeframe, inclusive, in which to return event volume. All
              datetime values are converted to UTC time. If the specified time isn't
              hour-aligned, the response includes the event volume count for the hour the time
              falls in.

          cursor: Cursor for pagination. This can be populated by the `next_cursor` value returned
              from the initial request.

          limit: The number of items to fetch. Defaults to 20.

          timeframe_end: The end of the timeframe, exclusive, in which to return event volume. If not
              specified, the current time is used. All datetime values are converted to UTC
              time.If the specified time isn't hour-aligned, the response includes the event
              volumecount for the hour the time falls in.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get(
            "/events/volume",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "timeframe_start": timeframe_start,
                        "cursor": cursor,
                        "limit": limit,
                        "timeframe_end": timeframe_end,
                    },
                    volume_list_params.VolumeListParams,
                ),
            ),
            cast_to=EventVolumes,
        )


class AsyncVolume(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncVolumeWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/orbcorp/orb-python#accessing-raw-response-data-eg-headers
        """
        return AsyncVolumeWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncVolumeWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/orbcorp/orb-python#with_streaming_response
        """
        return AsyncVolumeWithStreamingResponse(self)

    async def list(
        self,
        *,
        timeframe_start: Union[str, datetime],
        cursor: Optional[str] | Omit = omit,
        limit: int | Omit = omit,
        timeframe_end: Union[str, datetime] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> EventVolumes:
        """
        This endpoint returns the event volume for an account in a
        [paginated list format](/api-reference/pagination).

        The event volume is aggregated by the hour and the
        [timestamp](/api-reference/event/ingest-events) field is used to determine which
        hour an event is associated with. Note, this means that late-arriving events
        increment the volume count for the hour window the timestamp is in, not the
        latest hour window.

        Each item in the response contains the count of events aggregated by the hour
        where the start and end time are hour-aligned and in UTC. When a specific
        timestamp is passed in for either start or end time, the response includes the
        hours the timestamp falls in.

        Args:
          timeframe_start: The start of the timeframe, inclusive, in which to return event volume. All
              datetime values are converted to UTC time. If the specified time isn't
              hour-aligned, the response includes the event volume count for the hour the time
              falls in.

          cursor: Cursor for pagination. This can be populated by the `next_cursor` value returned
              from the initial request.

          limit: The number of items to fetch. Defaults to 20.

          timeframe_end: The end of the timeframe, exclusive, in which to return event volume. If not
              specified, the current time is used. All datetime values are converted to UTC
              time.If the specified time isn't hour-aligned, the response includes the event
              volumecount for the hour the time falls in.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._get(
            "/events/volume",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=await async_maybe_transform(
                    {
                        "timeframe_start": timeframe_start,
                        "cursor": cursor,
                        "limit": limit,
                        "timeframe_end": timeframe_end,
                    },
                    volume_list_params.VolumeListParams,
                ),
            ),
            cast_to=EventVolumes,
        )


class VolumeWithRawResponse:
    def __init__(self, volume: Volume) -> None:
        self._volume = volume

        self.list = _legacy_response.to_raw_response_wrapper(
            volume.list,
        )


class AsyncVolumeWithRawResponse:
    def __init__(self, volume: AsyncVolume) -> None:
        self._volume = volume

        self.list = _legacy_response.async_to_raw_response_wrapper(
            volume.list,
        )


class VolumeWithStreamingResponse:
    def __init__(self, volume: Volume) -> None:
        self._volume = volume

        self.list = to_streamed_response_wrapper(
            volume.list,
        )


class AsyncVolumeWithStreamingResponse:
    def __init__(self, volume: AsyncVolume) -> None:
        self._volume = volume

        self.list = async_to_streamed_response_wrapper(
            volume.list,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from ..._models import BaseModel

__all__ = ["PackageConfig"]


class PackageConfig(BaseModel):
    package_amount: str
    """A currency amount to rate usage by"""

    package_size: int
    """An integer amount to represent package size.

    For example, 1000 here would divide usage by 1000 before multiplying by
    package_amount in rating
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Optional
from typing_extensions import TypedDict

__all__ = ["AddressInputParam"]


class AddressInputParam(TypedDict, total=False):
    city: Optional[str]

    country: Optional[str]

    line1: Optional[str]

    line2: Optional[str]

    postal_code: Optional[str]

    state: Optional[str]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Union, Optional
from datetime import date, datetime
from typing_extensions import Annotated, TypedDict

from .._utils import PropertyInfo

__all__ = ["InvoiceUpdateParams"]


class InvoiceUpdateParams(TypedDict, total=False):
    due_date: Annotated[Union[Union[str, date], Union[str, datetime], None], PropertyInfo(format="iso8601")]
    """An optional custom due date for the invoice.

    If not set, the due date will be calculated based on the `net_terms` value.
    """

    metadata: Optional[Dict[str, Optional[str]]]
    """User-specified key/value pairs for the resource.

    Individual keys can be removed by setting the value to `null`, and the entire
    metadata mapping can be cleared by setting `metadata` to `null`.
    """

    net_terms: Optional[int]
    """The net terms determines the due date of the invoice.

    Due date is calculated based on the invoice or issuance date, depending on the
    account's configured due date calculation method. A value of '0' here represents
    that the invoice is due on issue, whereas a value of '30' represents that the
    customer has 30 days to pay the invoice. Do not set this field if you want to
    set a custom due date.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Iterable, Optional

import httpx

from ... import _legacy_response
from ..._types import Body, Omit, Query, Headers, NotGiven, omit, not_given
from ..._utils import maybe_transform, async_maybe_transform
from ..._compat import cached_property
from ..._resource import SyncAPIResource, AsyncAPIResource
from ..._response import to_streamed_response_wrapper, async_to_streamed_response_wrapper
from ...types.beta import external_plan_id_create_plan_version_params, external_plan_id_set_default_plan_version_params
from ...types.plan import Plan
from ..._base_client import make_request_options
from ...types.plan_version import PlanVersion

__all__ = ["ExternalPlanID", "AsyncExternalPlanID"]


class ExternalPlanID(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> ExternalPlanIDWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/orbcorp/orb-python#accessing-raw-response-data-eg-headers
        """
        return ExternalPlanIDWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> ExternalPlanIDWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/orbcorp/orb-python#with_streaming_response
        """
        return ExternalPlanIDWithStreamingResponse(self)

    def create_plan_version(
        self,
        external_plan_id: str,
        *,
        version: int,
        add_adjustments: Optional[Iterable[external_plan_id_create_plan_version_params.AddAdjustment]] | Omit = omit,
        add_prices: Optional[Iterable[external_plan_id_create_plan_version_params.AddPrice]] | Omit = omit,
        remove_adjustments: Optional[Iterable[external_plan_id_create_plan_version_params.RemoveAdjustment]]
        | Omit = omit,
        remove_prices: Optional[Iterable[external_plan_id_create_plan_version_params.RemovePrice]] | Omit = omit,
        replace_adjustments: Optional[Iterable[external_plan_id_create_plan_version_params.ReplaceAdjustment]]
        | Omit = omit,
        replace_prices: Optional[Iterable[external_plan_id_create_plan_version_params.ReplacePrice]] | Omit = omit,
        set_as_default: Optional[bool] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> PlanVersion:
        """
        This endpoint allows the creation of a new plan version for an existing plan.

        Args:
          version: New version number.

          add_adjustments: Additional adjustments to be added to the plan.

          add_prices: Additional prices to be added to the plan.

          remove_adjustments: Adjustments to be removed from the plan.

          remove_prices: Prices to be removed from the plan.

          replace_adjustments: Adjustments to be replaced with additional adjustments on the plan.

          replace_prices: Prices to be replaced with additional prices on the plan.

          set_as_default: Set this new plan version as the default

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not external_plan_id:
            raise ValueError(f"Expected a non-empty value for `external_plan_id` but received {external_plan_id!r}")
        return self._post(
            f"/plans/external_plan_id/{external_plan_id}/versions",
            body=maybe_transform(
                {
                    "version": version,
                    "add_adjustments": add_adjustments,
                    "add_prices": add_prices,
                    "remove_adjustments": remove_adjustments,
                    "remove_prices": remove_prices,
                    "replace_adjustments": replace_adjustments,
                    "replace_prices": replace_prices,
                    "set_as_default": set_as_default,
                },
                external_plan_id_create_plan_version_params.ExternalPlanIDCreatePlanVersionParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=PlanVersion,
        )

    def fetch_plan_version(
        self,
        version: str,
        *,
        external_plan_id: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> PlanVersion:
        """This API endpoint is in beta and its interface may change.

        It is recommended for
        use only in test mode.

        This endpoint is used to fetch a plan version. It returns the phases, prices,
        and adjustments present on this version of the plan.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not external_plan_id:
            raise ValueError(f"Expected a non-empty value for `external_plan_id` but received {external_plan_id!r}")
        if not version:
            raise ValueError(f"Expected a non-empty value for `version` but received {version!r}")
        return self._get(
            f"/plans/external_plan_id/{external_plan_id}/versions/{version}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=PlanVersion,
        )

    def set_default_plan_version(
        self,
        external_plan_id: str,
        *,
        version: int,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> Plan:
        """This API endpoint is in beta and its interface may change.

        It is recommended for
        use only in test mode.

        This endpoint allows setting the default version of a plan.

        Args:
          version: Plan version to set as the default.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not external_plan_id:
            raise ValueError(f"Expected a non-empty value for `external_plan_id` but received {external_plan_id!r}")
        return self._post(
            f"/plans/external_plan_id/{external_plan_id}/set_default_version",
            body=maybe_transform(
                {"version": version},
                external_plan_id_set_default_plan_version_params.ExternalPlanIDSetDefaultPlanVersionParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=Plan,
        )


class AsyncExternalPlanID(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncExternalPlanIDWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/orbcorp/orb-python#accessing-raw-response-data-eg-headers
        """
        return AsyncExternalPlanIDWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncExternalPlanIDWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/orbcorp/orb-python#with_streaming_response
        """
        return AsyncExternalPlanIDWithStreamingResponse(self)

    async def create_plan_version(
        self,
        external_plan_id: str,
        *,
        version: int,
        add_adjustments: Optional[Iterable[external_plan_id_create_plan_version_params.AddAdjustment]] | Omit = omit,
        add_prices: Optional[Iterable[external_plan_id_create_plan_version_params.AddPrice]] | Omit = omit,
        remove_adjustments: Optional[Iterable[external_plan_id_create_plan_version_params.RemoveAdjustment]]
        | Omit = omit,
        remove_prices: Optional[Iterable[external_plan_id_create_plan_version_params.RemovePrice]] | Omit = omit,
        replace_adjustments: Optional[Iterable[external_plan_id_create_plan_version_params.ReplaceAdjustment]]
        | Omit = omit,
        replace_prices: Optional[Iterable[external_plan_id_create_plan_version_params.ReplacePrice]] | Omit = omit,
        set_as_default: Optional[bool] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> PlanVersion:
        """
        This endpoint allows the creation of a new plan version for an existing plan.

        Args:
          version: New version number.

          add_adjustments: Additional adjustments to be added to the plan.

          add_prices: Additional prices to be added to the plan.

          remove_adjustments: Adjustments to be removed from the plan.

          remove_prices: Prices to be removed from the plan.

          replace_adjustments: Adjustments to be replaced with additional adjustments on the plan.

          replace_prices: Prices to be replaced with additional prices on the plan.

          set_as_default: Set this new plan version as the default

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not external_plan_id:
            raise ValueError(f"Expected a non-empty value for `external_plan_id` but received {external_plan_id!r}")
        return await self._post(
            f"/plans/external_plan_id/{external_plan_id}/versions",
            body=await async_maybe_transform(
                {
                    "version": version,
                    "add_adjustments": add_adjustments,
                    "add_prices": add_prices,
                    "remove_adjustments": remove_adjustments,
                    "remove_prices": remove_prices,
                    "replace_adjustments": replace_adjustments,
                    "replace_prices": replace_prices,
                    "set_as_default": set_as_default,
                },
                external_plan_id_create_plan_version_params.ExternalPlanIDCreatePlanVersionParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=PlanVersion,
        )

    async def fetch_plan_version(
        self,
        version: str,
        *,
        external_plan_id: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> PlanVersion:
        """This API endpoint is in beta and its interface may change.

        It is recommended for
        use only in test mode.

        This endpoint is used to fetch a plan version. It returns the phases, prices,
        and adjustments present on this version of the plan.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not external_plan_id:
            raise ValueError(f"Expected a non-empty value for `external_plan_id` but received {external_plan_id!r}")
        if not version:
            raise ValueError(f"Expected a non-empty value for `version` but received {version!r}")
        return await self._get(
            f"/plans/external_plan_id/{external_plan_id}/versions/{version}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=PlanVersion,
        )

    async def set_default_plan_version(
        self,
        external_plan_id: str,
        *,
        version: int,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> Plan:
        """This API endpoint is in beta and its interface may change.

        It is recommended for
        use only in test mode.

        This endpoint allows setting the default version of a plan.

        Args:
          version: Plan version to set as the default.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not external_plan_id:
            raise ValueError(f"Expected a non-empty value for `external_plan_id` but received {external_plan_id!r}")
        return await self._post(
            f"/plans/external_plan_id/{external_plan_id}/set_default_version",
            body=await async_maybe_transform(
                {"version": version},
                external_plan_id_set_default_plan_version_params.ExternalPlanIDSetDefaultPlanVersionParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=Plan,
        )


class ExternalPlanIDWithRawResponse:
    def __init__(self, external_plan_id: ExternalPlanID) -> None:
        self._external_plan_id = external_plan_id

        self.create_plan_version = _legacy_response.to_raw_response_wrapper(
            external_plan_id.create_plan_version,
        )
        self.fetch_plan_version = _legacy_response.to_raw_response_wrapper(
            external_plan_id.fetch_plan_version,
        )
        self.set_default_plan_version = _legacy_response.to_raw_response_wrapper(
            external_plan_id.set_default_plan_version,
        )


class AsyncExternalPlanIDWithRawResponse:
    def __init__(self, external_plan_id: AsyncExternalPlanID) -> None:
        self._external_plan_id = external_plan_id

        self.create_plan_version = _legacy_response.async_to_raw_response_wrapper(
            external_plan_id.create_plan_version,
        )
        self.fetch_plan_version = _legacy_response.async_to_raw_response_wrapper(
            external_plan_id.fetch_plan_version,
        )
        self.set_default_plan_version = _legacy_response.async_to_raw_response_wrapper(
            external_plan_id.set_default_plan_version,
        )


class ExternalPlanIDWithStreamingResponse:
    def __init__(self, external_plan_id: ExternalPlanID) -> None:
        self._external_plan_id = external_plan_id

        self.create_plan_version = to_streamed_response_wrapper(
            external_plan_id.create_plan_version,
        )
        self.fetch_plan_version = to_streamed_response_wrapper(
            external_plan_id.fetch_plan_version,
        )
        self.set_default_plan_version = to_streamed_response_wrapper(
            external_plan_id.set_default_plan_version,
        )


class AsyncExternalPlanIDWithStreamingResponse:
    def __init__(self, external_plan_id: AsyncExternalPlanID) -> None:
        self._external_plan_id = external_plan_id

        self.create_plan_version = async_to_streamed_response_wrapper(
            external_plan_id.create_plan_version,
        )
        self.fetch_plan_version = async_to_streamed_response_wrapper(
            external_plan_id.fetch_plan_version,
        )
        self.set_default_plan_version = async_to_streamed_response_wrapper(
            external_plan_id.set_default_plan_version,
        )
